@@ -1,5 +1,6 @@
 /* coff object file format
-   Copyright (C) 1989, 90, 91, 92, 94, 1995 Free Software Foundation, Inc.
+   Copyright (C) 1989, 90, 91, 92, 94, 95, 96, 97, 1998
+   Free Software Foundation, Inc.
 
    This file is part of GAS.
 
@@ -14,8 +15,9 @@
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
-   along with GAS; see the file COPYING.  If not, write to
-   the Free Software Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */
+   along with GAS; see the file COPYING.  If not, write to the Free
+   Software Foundation, 59 Temple Place - Suite 330, Boston, MA
+   02111-1307, USA.  */
 
 #ifndef OBJ_FORMAT_H
 #define OBJ_FORMAT_H
@@ -61,11 +63,6 @@
 
 #ifdef TC_SPARC
 #include "coff/sparc.h"
-#ifdef TE_LYNX
-#define TARGET_FORMAT "coff-sparc-lynx"
-#else
-#define TARGET_FORMAT "coff-sparc"
-#endif
 #endif
 
 #ifdef TC_I386
@@ -114,7 +111,10 @@
 
 #ifdef TC_SH
 #include "coff/sh.h"
-#define TARGET_FORMAT (shl ? "coff-shl" : "coff-sh")
+#define TARGET_FORMAT					\
+  (shl							\
+   ? (sh_small ? "coff-shl-small" : "coff-shl")		\
+   : (sh_small ? "coff-sh-small" : "coff-sh"))
 #endif
 
 #ifdef TC_M88K
@@ -127,6 +127,12 @@
 #define TARGET_FORMAT "coff-w65"
 #endif
 
+#ifdef TC_TIC30
+#include "coff/tic30.h"
+#define TARGET_FORMAT "coff-tic30"
+#endif
+
+
 /* Targets may also set this.  Also, if BFD_ASSEMBLER is defined, this
    will already have been defined.  */
 #undef SYMBOLS_NEED_BACKPOINTERS
@@ -136,6 +142,12 @@
 #define OBJ_COFF_MAX_AUXENTRIES 1
 #endif /* OBJ_COFF_MAX_AUXENTRIES */
 
+extern void coff_obj_symbol_new_hook PARAMS ((struct symbol *));
+#define obj_symbol_new_hook coff_obj_symbol_new_hook
+
+extern void coff_obj_read_begin_hook PARAMS ((void));
+#define obj_read_begin_hook coff_obj_read_begin_hook
+
 /* ***********************************************************************
 
    This file really contains two implementations of the COFF back end.
@@ -197,7 +209,6 @@
 #define SYM_AUXENT(S)	(&coffsymbol ((S)->bsym)->native[1].u.auxent)
 
 #define DO_NOT_STRIP	0
-#define DO_STRIP	1
 
 extern void obj_coff_section PARAMS ((int));
 
@@ -332,11 +343,11 @@ extern void coff_frob_symbol PARAMS ((struct symbol *, int *));
 extern void coff_adjust_symtab PARAMS ((void));
 extern void coff_frob_section PARAMS ((segT));
 extern void coff_adjust_section_syms PARAMS ((bfd *, asection *, PTR));
-extern void coff_frob_file PARAMS ((void));
+extern void coff_frob_file_after_relocs PARAMS ((void));
 #define obj_frob_symbol(S,P) 	coff_frob_symbol(S,&P)
 #define obj_adjust_symtab()	coff_adjust_symtab()
 #define obj_frob_section(S)	coff_frob_section (S)
-#define obj_frob_file()		coff_frob_file ()
+#define obj_frob_file_after_relocs() coff_frob_file_after_relocs ()
 
 extern struct symbol *coff_last_function;
 
@@ -402,7 +413,6 @@ typedef struct
 
 #ifndef DO_NOT_STRIP
 #define DO_NOT_STRIP	0
-#define DO_STRIP	1
 #endif
 /* Symbol table macros and constants */
 
@@ -415,7 +425,7 @@ typedef struct
 #define C_DEBUG_SECTION		N_DEBUG
 #define C_NTV_SECTION		N_TV
 #define C_PTV_SECTION		P_TV
-#define C_REGISTER_SECTION	20
+#define C_REGISTER_SECTION	50
 
 /*
  *  Macros to extract information from a symbol table entry.
@@ -440,8 +450,12 @@ typedef struct
 /* A symbol name whose name includes ^A is a gas internal pseudo symbol */
 #define S_IS_LOCAL(s) \
   ((s)->sy_symbol.ost_entry.n_scnum == C_REGISTER_SECTION \
-   || (S_LOCAL_NAME(s) && !flag_keep_locals) \
-   || (strchr (S_GET_NAME (s), '\001') != NULL))
+   || (S_LOCAL_NAME(s) && ! flag_keep_locals && ! S_IS_DEBUG (s)) \
+   || strchr (S_GET_NAME (s), '\001') != NULL \
+   || strchr (S_GET_NAME (s), '\002') != NULL \
+   || (flag_strip_local_absolute \
+       && !S_IS_EXTERNAL(s) \
+       && (s)->sy_symbol.ost_entry.n_scnum == C_ABS_SECTION))
 /* True if a symbol is not defined in this file */
 #define S_IS_EXTERN(s)		((s)->sy_symbol.ost_entry.n_scnum == 0 \
 				 && S_GET_VALUE (s) == 0)
@@ -551,6 +565,7 @@ typedef struct
 #define SF_TAG		(0x00080000)	/* Is a tag */
 #define SF_DEBUG	(0x00100000)	/* Is in debug or abs section */
 #define SF_GET_SEGMENT	(0x00200000)	/* Get the section of the forward symbol. */
+#define SF_ADJ_LNNOPTR	(0x00400000)	/* Has a lnnoptr */
 /* All other bits are unused. */
 
 /* Accessors */
@@ -568,6 +583,7 @@ typedef struct
 #define SF_GET_TAGGED(s)	(SF_GET (s) & SF_TAGGED)
 #define SF_GET_TAG(s)		(SF_GET (s) & SF_TAG)
 #define SF_GET_GET_SEGMENT(s)	(SF_GET (s) & SF_GET_SEGMENT)
+#define SF_GET_ADJ_LNNOPTR(s)	(SF_GET (s) & SF_ADJ_LNNOPTR)
 #define SF_GET_I960(s)		(SF_GET (s) & SF_I960_MASK)	/* used by i960 */
 #define SF_GET_BALNAME(s)	(SF_GET (s) & SF_BALNAME)	/* used by i960 */
 #define SF_GET_CALLNAME(s)	(SF_GET (s) & SF_CALLNAME)	/* used by i960 */
@@ -590,6 +606,7 @@ typedef struct
 #define SF_SET_TAGGED(s)	(SF_GET (s) |= SF_TAGGED)
 #define SF_SET_TAG(s)		(SF_GET (s) |= SF_TAG)
 #define SF_SET_GET_SEGMENT(s)	(SF_GET (s) |= SF_GET_SEGMENT)
+#define SF_SET_ADJ_LNNOPTR(s)	(SF_GET (s) |= SF_ADJ_LNNOPTR)
 #define SF_SET_I960(s,v)	(SF_GET (s) |= ((v) & SF_I960_MASK))	/* used by i960 */
 #define SF_SET_BALNAME(s)	(SF_GET (s) |= SF_BALNAME)	/* used by i960 */
 #define SF_SET_CALLNAME(s)	(SF_GET (s) |= SF_CALLNAME)	/* used by i960 */
@@ -769,8 +786,25 @@ extern struct internal_scnhdr text_section_header;
    ? (S_SET_SEGMENT (dest, S_GET_SEGMENT (src)), 0) \
    : 0)
 
+#ifdef TE_PE
+#define obj_handle_link_once(t) obj_coff_pe_handle_link_once (t)
+extern void obj_coff_pe_handle_link_once ();
+#endif
+
 #endif /* not BFD_ASSEMBLER */
 
+/* In COFF, if a symbol is defined using .def/.val SYM/.endef, it's OK
+   to redefine the symbol later on.  This can happen if C symbols use
+   a prefix, and a symbol is defined both with and without the prefix,
+   as in start/_start/__start in gcc/libgcc1-test.c.  */
+#define RESOLVE_SYMBOL_REDEFINITION(sym)		\
+(SF_GET_GET_SEGMENT (sym)				\
+ ? (sym->sy_frag = frag_now,				\
+    S_SET_VALUE (sym, frag_now_fix ()),			\
+    S_SET_SEGMENT (sym, now_seg),			\
+    0)							\
+ : 0)
+
 /* Stabs in a coff file go into their own section.  */
 #define SEPARATE_STAB_SECTIONS 1
 
